/* ===============================================================
  Project: Gestion électrique d'une boucle de retournement à 2 voies
  Author: Philippe CHAVATTE
  Web: http://lormedy.free.fr
  Created: 24/06/2023
  Last modified : 07/10/2023
  Target : Arduino NANO V3 328P-32TQFP (old bootloader)
  Arduino IDE: 1.8.19
  Utilisation de la bibliothèque VL6180X_WE version 1.1.5 dans le dossier: C:\Users\...\Documents\Arduino\libraries\VL6180X_WE 
  Utilisation de la bibliothèque Wire version 1.0 dans le dossier: C:\Users\...\Documents\Arduino\libraries\Wire 
  Le croquis utilise 8814 octets (28%) de l'espace de stockage de programmes. Le maximum est de 30720 octets.
  Les variables globales utilisent 562 octets (27%) de mémoire dynamique, ce qui laisse 1486 octets pour les variables locales. Le maximum est de 2048 octets.

  Description: Controle d'inversion électrique DCC pour anneau de retournement sur double voies avec 
             2 distance detectors VL6180X using I2C and interruptions
             to 4 x 5V Relay Modules output for double "boucle de retournement"
             Two contacts are used per track, so we use 2 relays for one track and two for the opposite track

  Info : 190 km/h ~ 60 cm/s at 1/87 scale
         160 km/h ~ 51 cm/s at 1/87 scale
          60 km/h ~ 19 cm/s at 1/87 scale
          H0 2.75 m at 60 cm/s ~ 4.6 s
          H0 2.75 m at 51 cm/s ~ 5.4 s

  Connect 5VDC on Arduino to VCC on Relay Module
  Connect GND on Arduino to GND on Relay Module
  Connect +5VDC and GND to the 2 IR detectors
  Connect 2 VL6180X detectors to Arduino inputs, à la limite électrique du retournement
  Connect 2 Arduino outputs to Relays Module, 2 pour commuter chaque retournement
  Connect tracks to the Common Terminal (middle terminal) on Relays Module.
  Connect DCC to the NO/NC Terminals (outer terminal) on Relays Module.
  Connect +5VDC, GND, ,I2C and Int to the 2 VL6180X detectors

  D2  in  int GPIO capteur 14
  D3  in  int GPIO capteur 32
  D4 out relai 14 & S88 39 ou 55
  D5 out relai 32 & S88 40 ou 56

  A2  in  ShutDown 14
  A3  in  ShutDown 32
  A4  in  SDA
  A5  out SCL
================================================================== */

#include <VL6180X_WE.h>  // Bibliothèque modifiée pour gérer les interruptions la distance limite de détection

// ========================================================
// Constantes
// ========================================================
// distance de détection à exclure
#define LIMITE_DETECTION_14  199   // limite de détection 14
#define LIMITE_DETECTION_32  199   // limite de détection 32
#define DISTANCE_MAXI  255

// address we will assign if dual sensor is present
#define VL6180X_ADDRESS 0x29        // default address
#define VL6180X_14_ADDRESS 0x14
#define VL6180X_32_ADDRESS 0x32

// objects for each VL6180X sensor
VL6180x sensor(VL6180X_ADDRESS);
VL6180x sensor14(VL6180X_14_ADDRESS);
VL6180x sensor32(VL6180X_32_ADDRESS);

#define DELAI 500  // retard pour libèrer le relai

// external pins connection
// Cablage avec Time-of-Flight IR
//      pin name            pin number                  connector
#define TOF_14_pin          2   // Connect D2   Digital Pin 5 on Arduino Nano to outer Time-of-Flight IR Detector interrupt
#define TOF_32_pin          3   // Connect D3   Digital Pin 6 on Arduino Nano to inner Time-of-Flight IR Detector interrupt
#define RELAY_14_pin        4   // Connect D4   Digital Pin 7 on Arduino Nano to CH1 and CH2 on Relay Module (outer)
#define RELAY_32_pin        5   // Connect D5   Digital Pin 8 on Arduino Nano to CH3 and CH4 on Relay Module (inner)

// pins for shutdown
#define SHT_VL6180X_14_pin A2
#define SHT_VL6180X_32_pin A3

//#define SDA                A4   // Connect A4   Analog  Pin J1 30 on Arduino to I2C SDA
//#define SCL                A5   // Connect A5   Analog  Pin J1 31 on Arduino to I2C SCL

//#define DEBUG              // enable display
#define LOG14
#define LOG32

// ========================================================
// Variables
// ========================================================
volatile bool event14 = false;
volatile bool event32 = false;
uint8_t distance14;
uint8_t distance32;

volatile boolean VL6180X_14 = 0;
volatile boolean VL6180X_32 = 0;
boolean VL6180X_14_z = 0;
boolean VL6180X_32_z = 0;
boolean VL6180X_14_occupe = 0;
boolean VL6180X_32_occupe = 0;

boolean R14 = false;
boolean R32 = false;
boolean R14_mem = true;
boolean R32_mem = true;

uint32_t maintenant;
uint32_t delai14;
uint32_t delai32;

// ========================================================
// Initialyze both VL6180X
// ========================================================
void initTOF() {
  // all reset
  digitalWrite(SHT_VL6180X_14_pin, LOW);
  digitalWrite(SHT_VL6180X_32_pin, LOW);
  pinMode(SHT_VL6180X_14_pin, OUTPUT);
  pinMode(SHT_VL6180X_32_pin, OUTPUT);
  Serial.println("Shutdown pins inited...");
  delay(500);

  // activating VL6180X_14 and reseting VL6180X_32
  pinMode(SHT_VL6180X_14_pin, INPUT);
  pinMode(SHT_VL6180X_32_pin, OUTPUT);
  delay(500);

  // initialising VL6180X_14
  if(sensor.VL6180xInit() != 0) {
    Serial.println(F("sensor14 FAILED TO INITIALIZE")); //Initialize device and check for errors
  }
  else {
    sensor.changeAddress(VL6180X_ADDRESS, VL6180X_14_ADDRESS);
    delay(100);
    sensor14.VL6180xDefautSettings(); //Load default settings to get started.
    delay(100);
    // Range Threshold Interrupt: VL6180xSetDistInt(lower limit, upper limit)
    sensor14.VL6180xSetDistInt(LIMITE_DETECTION_14,DISTANCE_MAXI);  // inhibe les interruptions si distance > 200mm
    //Get Distance and report in mm
    Serial.print("Distance14 measured (mm) = ");
    Serial.println( sensor14.getDistance() ); 
    sensor14.getDistanceContinously();
  }
  delay(100);

  // activating VL6180X_32
  pinMode(SHT_VL6180X_32_pin, INPUT);
  delay(500);

  // initing VL6180X_32
  if(sensor.VL6180xInit() != 0) {
    Serial.println(F("sensor32 FAILED TO INITIALIZE"));             //Initialize device and check for errors
  }
  else {
    sensor.changeAddress(VL6180X_ADDRESS, VL6180X_32_ADDRESS);
    delay(100);
    sensor32.VL6180xDefautSettings(); //Load default settings to get started.
    delay(100);
    sensor32.VL6180xSetDistInt(LIMITE_DETECTION_32,DISTANCE_MAXI);  // inhibe les interruptions si distance > 200mm
    //Get Distance and report in mm
    Serial.print("Distance32 measured (mm) = ");
    Serial.println( sensor32.getDistance() ); 
    sensor32.getDistanceContinously();
  }
  delay(500); // delay 0.5 second before starting operations
  Serial.println("Init done...");
}

// ========================================================
// Interrupts
// ========================================================
void ISR14(){
  event14 = true;
}

void ISR32(){
  event32 = true;
}

// ========================================================
// Initialisation du Nano
// ========================================================
void setup() {
  //Setup all the Arduino Pins
  pinMode(TOF_14_pin, INPUT);           // Low active by TOF IR detection
  pinMode(TOF_32_pin, INPUT);

  Serial.println(F("Relais désactivés"));
  //Turn OFF any power to the Relay channels
  digitalWrite(RELAY_14_pin, HIGH);   //relay off
  digitalWrite(RELAY_32_pin, HIGH);   //relay off
  delay(100);
  pinMode(RELAY_14_pin, OUTPUT);
  pinMode(RELAY_32_pin, OUTPUT);

  pinMode(LED_BUILTIN, OUTPUT);

  //Initialize the serial port
  Serial.begin(115200);
  Serial.println();
  Serial.println(F("Module de gestion de boucle de retournement à double voie - version 07/10/2023 TOF"));
  Serial.println(F("Initialisation de la liaison série via USB à 115200 bauds"));

  // TOF sensors
  Wire.begin(); //Start I2C library
  Serial.println(F("TOF VL6280X WE with interrupt"));
  Serial.print("A2 "); Serial.print(analogRead(A2)); Serial.print(" - ");
  Serial.print(analogRead(A2)*5/1.024); Serial.println(" mV");  
  Serial.print("A3 "); Serial.print(analogRead(A3)); Serial.print(" - ");
  Serial.print(analogRead(A3)*5/1.024); Serial.println(" mV");  

  initTOF();

  attachInterrupt(digitalPinToInterrupt(TOF_14_pin), ISR14, FALLING);
  attachInterrupt(digitalPinToInterrupt(TOF_32_pin), ISR32, FALLING);
}

// ========================================================
// Main loop
// ========================================================
void loop() {
  maintenant = millis();
  if(event14){
    //Serial.println(F("Interrupt 14 !"));

    distance14 = sensor14.getLastDistanceFromHistory();
    if ((distance14 > 30) && (distance14 < LIMITE_DETECTION_14)) {
      VL6180X_14 = true;
      Serial.print(F("Last Distance14 Value: "));
      Serial.print(distance14);
      Serial.print(F("\t /  "));
      Serial.print(VL6180X_14_occupe);
      Serial.print(F(" / Relai 14 " ));
      Serial.print(R14==false ? "Off": "On");
/*    
      if (distance14<10) Serial.print("     ");
      else if (distance14<100) Serial.print("    ");
      else Serial.print("   ");
      for (uint8_t i=0; i<distance14; i++) {
        Serial.print("*");
      }
*/
      Serial.println();
    }
    digitalWrite(LED_BUILTIN, HIGH);
    event14 = false;
    sensor14.VL6180xClearInterrupt();
    delai14 = maintenant;
  }

  if(event32){
    //Serial.println(F("Interrupt 32 !"));
    distance32 = sensor32.getLastDistanceFromHistory();
    if ((distance32 > 30) && (distance32 < LIMITE_DETECTION_32)) {
      VL6180X_32 = true;
      Serial.print(F("Last Distance32 Value: "));
      Serial.print(distance32);
      Serial.print(F("\t /  " ));
      Serial.print(VL6180X_32_occupe);
      Serial.print(F(" / Relai 32 "));
      Serial.print(R32==false ? "Off": "On");
/*
      if (distance32<10) Serial.print("     ");
      else if (distance32<100) Serial.print("    ");
      else Serial.print("   ");
      for (uint8_t i=0; i<distance32; i++) {
        Serial.print("*");
      }
*/
      Serial.println();
    }
    digitalWrite(LED_BUILTIN, HIGH);
    event32 = false;
    sensor32.VL6180xClearInterrupt();
    delai32 = maintenant;
  }

  VL6180X_14 &= (distance14 < DISTANCE_MAXI);
  VL6180X_32 &= (distance32 < DISTANCE_MAXI);
  
// Mémorisation des capteurs TOF
  VL6180X_14_occupe = VL6180X_14 || VL6180X_14_occupe;
  VL6180X_32_occupe = VL6180X_32 || VL6180X_32_occupe;

  // ********** Commande du basculement des relais **********

  //  Controle du retournement de Voie_1 vers Voie_4 ou de Voie_4 vers Voie_1
  if (VL6180X_14_occupe) {
    digitalWrite(RELAY_14_pin, LOW);     //Relay on, Relay 14 switches to NO      (inversion)
    R14 = true;
  }
  else {
    digitalWrite(RELAY_14_pin, HIGH);    //Relay off, Relay 14 switches to NC     (normal)
    R14 = false;
  }

  // Controle du retournement de Voie_2 vers Voie_3 ou de Voie_3 vers Voie_2
  if (VL6180X_32_occupe) {
    digitalWrite(RELAY_32_pin, LOW);     //Relay on, Relay 32 switches to NO      (inversion)
    R32 = true;
  }
  else {
    digitalWrite(RELAY_32_pin, HIGH);    //Relay off, Relay 32 switches to NC     (normal)
    R32 = false;
  }

#ifdef LOG14
  if (R14_mem != R14) {
    Serial.print(F("RELAY_14  "));
    Serial.println(R14==false ? "Off": "On");
  }
  R14_mem = R14;
#endif
#ifdef LOG32
  if (R32_mem != R32) {
    Serial.print(F("RELAY_32  "));
    Serial.println(R32==false ? "Off": "On");
  }
  R32_mem = R32;
#endif

  // ********** print log

#if defined(DEBUG) // for test only
    Serial.print(F("   VL6180X_14 "));  Serial.print(VL6180X_14);
    Serial.print(F("   VL6180X_14z ")); Serial.print(VL6180X_14_z);
    Serial.print(F("   R14 "));     Serial.print(R14);
    Serial.print(F("   IRD32 "));   Serial.print(VL6180X_32);
    Serial.print(F("   IRD32z "));  Serial.println(VL6180X_32_z);
    Serial.print(F("   R32 "));     Serial.print(R32);
#endif

//  Serial.print(F("V14    "));   Serial.print(VL6180X_14);
//  Serial.print(F("    "));   Serial.print(VL6180X_14_z);
//  Serial.print(F("    "));   Serial.print(VL6180X_14_occupe);
//  Serial.println();
//  Serial.print(F("V32    "));   Serial.print(VL6180X_32);
//  Serial.print(F("    "));   Serial.print(VL6180X_32_z);
//  Serial.print(F("    "));   Serial.print(VL6180X_32_occupe);
//  Serial.println();

  // Libération des infos capteurs mémorisés donc des relais au prochain passage dans loop
  if (!VL6180X_14 && !VL6180X_14_z) VL6180X_14_occupe = false;
  if (!VL6180X_32 && !VL6180X_32_z) VL6180X_32_occupe = false;

  // Mémorisation de l'état précédent
  VL6180X_14_z = VL6180X_14;
  VL6180X_32_z = VL6180X_32;

  // clear lecture du capteur TOF après un délai
  if (maintenant > (delai14 + DELAI)) VL6180X_14 = false;
  if (maintenant > (delai32 + DELAI)) VL6180X_32 = false;

  delay(100); // Wait 100 milliseconds   // attendre 100ms pour filtrer les mauvais contacts rail/roue
  digitalWrite(LED_BUILTIN, LOW);
}
